/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */

package fr.amap.lidar.amapvox.visualization.engine3d.util;

import fr.amap.commons.util.swing.jprogress.JProgressEvent;
import fr.amap.commons.util.swing.jprogress.JProgressListener;
import fr.amap.commons.util.swing.jprogress.JProgressable;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;

/**
 * Voxel space file parsing class.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class VoxelSpaceFile extends JProgressable {
    
    private final List<String> names;
    private final Map<String, Float> max;
    private final Map<String, Float> min;
    private final Map<String, List<Float>> sortedPropertyValues;
    private final List<List<Float>> lines;
    private int width, height, depth;
    private String path;
    
    public VoxelSpaceFile() {
        super();
        this.names = new CopyOnWriteArrayList();
        this.lines = new CopyOnWriteArrayList();
        this.max = new HashMap();
        this.min = new HashMap();
        this.width = 0;
        this.height = 0;
        this.depth = 0;
        this.sortedPropertyValues = new HashMap();
    }

    public void setPath(final String path) {
        this.path = path;
    }

    public String getPath() {
        return path;
    }
    
    public void setSize(int width, int height, int depth) {
        this.width = width;
        this.height = height;
        this.depth = depth;
    }
    
    public int getWidth() {
        return width;
    }
    
    public int getHeight() {
        return height;
    }
    
    public int getDepth() {
        return depth;
    }
    
    public final void addName(final String name) {
        names.add(name);
    }
    
    public final void addLine(final List<Float> line) {
        lines.add(line);
    }
    
    public String[] getNames() {
        String[] array = new String[names.size()];
        
        for (int i = 0; i < array.length; i++) {
            array[i] = names.get(i);
        }
        
        return array;
    }
    
    public final Float getProperty(String name, int index) {
        if (lines.size() > index) {
            if (names.contains(name)) {
                return lines.get(index).get(names.indexOf(name));
            }
        }
        
        return null;
    }
    
    /**
     * Create a list of singles values of a property and sort them by size (lesser to greater).
     * @param property 
     */
    private void sortPropertyValues(final String property) {
        if (!sortedPropertyValues.containsKey(property)) {
            final List<Float> sortedValues = new ArrayList();

            for (List<Float> line : lines) {
                final float v = line.get(names.indexOf(property));

                if (!Float.isNaN(v)) {
                    if (!sortedValues.contains(v)) {
                        sortedValues.add(v);
                    }
                }
            }

            Collections.sort(sortedValues);
            sortedPropertyValues.put(property, sortedValues);
        }
    }
    
    /**
     * Get the number of singles values of a particular property.
     * @param property
     * @return 
     */
    public final int getDifferentValueQuantity(final String property) {
        sortPropertyValues(property);
        return sortedPropertyValues.get(property).size();
    }
    
    public final float getSingleProperty(final String property, final int index) {
        sortPropertyValues(property);
        return sortedPropertyValues.get(property).get(index);
    }
    
    /**
     * Get the closest existing value of an approximative 
     * and not existing property value.
     * @param property
     * @param value
     * @return 
     */
    public final float getClosestValue(final String property, final float value) {
        sortPropertyValues(property);
        
        float prev = Float.NEGATIVE_INFINITY;
        float closest = value;
        
        for (Float v : sortedPropertyValues.get(property)) {
            if (v != Float.NaN) {
                if (value == v) {
                    closest = v;
                } else if (value > prev && value < v) {
                    if (v - value > value - prev) {
                        closest = prev;
                    } else {
                        closest = v;
                    }
                }

                prev = v;
            }
        }
        
        return closest;
    }
    
    public final float getMinPropertyValue(String name) {
        if (min.isEmpty()) {
            throw(new NullPointerException(
                    "The " + name + " property doesn't contains any values."));
        }
        
        return min.get(name);
    }
    
    public final float getMaxPropertyValue(String name) {    
        if (max.isEmpty()) {
            throw(new NullPointerException(
                    "The " + name + " property doesn't contains any values."));
        }
        
        return max.get(name);
    }
    
    public final List<Float> getLine(String name) {
        List<Float> values = new ArrayList<>();
        
        for (int i = 0; i < lines.size(); i++) {
            values.add(getProperty(name, i));
        }
        
        return values;
    }
    
    public final boolean parse(final InputStream voxelFile) {                      
        if (voxelFile != null) {
            names.clear();
            min.clear();
            max.clear();
            lines.clear();
                        
            try {
                final BufferedReader br = new BufferedReader(new InputStreamReader(voxelFile));
                                
                if (br.ready()) {
                    for (JProgressListener progressListener : getProgressListeners()) {
                        progressListener.onProgress(new JProgressEvent(
                                1, 3, 
                                "Check if it's a Voxel Space file..."));
                    }
                    
                    if (br.readLine().contains("VOXEL SPACE")) {                        
                        String line;
                        
                        for (JProgressListener progressListener : getProgressListeners()) {
                            progressListener.onProgress(new JProgressEvent(
                                    2, 3, 
                                    "Start parsing file..."));
                        }
                        
                        while ((line = br.readLine()) != null) {                            
                            if (!line.startsWith("#") && !line.startsWith("VOXEL SPACE")) {
                                if (line.startsWith("i j k")) {
                                    for (String name : line.split(" ")) {
                                        addName(name);
                                    }
                                } else {
                                    List<Float> values = new ArrayList();
                                    
                                    for (String value : line.split(" ")) {
                                        try {
                                            values.add(Float.valueOf(value));
                                        } catch (NumberFormatException ex) {
                                            values.add(Float.NaN);
                                        }
                                    }
                                    
                                    addLine(values);
                                }
                            } else if (line.startsWith("#split")) {
                                String[] values = line.substring(8).split(" ");
                                
                                setSize(Integer.parseInt(values[0]),
                                        Integer.parseInt(values[1]),
                                        Integer.parseInt(values[2]));
                            }
                        }
                        
                        for (JProgressListener progressListener : getProgressListeners()) {
                            progressListener.onProgress(new JProgressEvent(
                                    3, 3, 
                                    "Set min and max values..."));
                        }
                        
                        names.forEach((String name) -> {
                            for (int i = 0; i < lines.size(); i++) {
                                float value = getProperty(name, i);

                                if (!Float.isNaN(value)) {
                                    if (!max.containsKey(name)) {
                                        max.put(name, value);
                                    } else if (value > max.get(name)) {
                                        max.put(name, value);
                                    }
                                    
                                    if (!min.containsKey(name)) {
                                        min.put(name, value);
                                    } else if (value < min.get(name)) {
                                        min.put(name, value);
                                    }
                                }
                            }
                        });
                        
                        if (max.isEmpty() || min.isEmpty()) {
                            return false;
                        }
                    } else {
                        return false;
                    }
                    
                    br.close();
               }
            } catch (IOException ex) {
                ex.printStackTrace();
                return false;
            }
        } else {
            return false;
        }
        
        return true;
    }
    
}
